// $URL: http://subversion:8080/svn/gsc/trunk/drivers/gsc_common/driver/intime/os_irq.c $
// $Rev: 43205 $
// $Date: 2018-07-08 17:01:59 -0500 (Sun, 08 Jul 2018) $

// INtime: Device Driver: source file

#include "main.h"

#ifdef DEV_SUPPORTS_IRQ



// #define ********************************************************************

#define	PREFIX		DEV_NAME ": IRQ"



//*****************************************************************************
static __INTERRUPT void _irq_isr(WORD wCSRA, WORD wLevel, PVOID pv)
{
	__SHARED_INTERRUPT_PROLOG();
	EnterRtInterrupt(wLevel);

	// This device owns this interrupt.
	// Disable the interrupt then signal the thread.
	// The thread will subsequently enable the interrupt.
	DisableRtInterrupt(wLevel);
	SignalRtInterruptThread(wLevel);

	__SHARED_INTERRUPT_RETURN();
}



//*****************************************************************************
static int _release_irq_sem(dev_data_t* dev)
{
	int	ret;

	ret	= os_sem_unlock(&dev->irq.os.sem);

	if (ret)
	{
		db_printf("%s: Failed to release IRQ semaphore.\n", PREFIX);

		if (dev->irq.os.ret == 0)
			dev->irq.os.ret	= ret;
	}

	return(ret);
}



//*****************************************************************************
static void _irq_thread(LPVOID arg)
{
	dev_data_t*	dev;
	WORD		irq_level;
	int			ret;
	BOOLEAN		sts;
	WORD		wd;


	dev					= (void*) arg;
	dev->irq.os.started	= 1;
	irq_level			= dev->irq.os.irq_level;

	for (;;)	// A convenience loop
	{
		// Install the ISR, making us the interrupt service thread.
		wd	= SetRtInterruptHandlerEx(irq_level | SHARED_LEVEL,	(BYTE) 255, _irq_isr, dev);

		if (wd == BAD_LEVEL)
		{
			// ISR installation failed.
			dev->irq.os.ret	= os_get_error_status();
			db_printf("%s: Failed to install ISR.\n", PREFIX);

			// Release the parent thread.
			ret	= _release_irq_sem(dev);
			dev->irq.os.failed	= __LINE__;
			break;
		}

		dev->irq.os.installed	= 1;
		dev->irq.os.irq_level	= wd;
		irq_level				= wd;

		// The interrupt is now enabled.

		// Release the parent thread.
		ret	= _release_irq_sem(dev);

		if (ret)
		{
			dev->irq.os.failed	= __LINE__;
			break;
		}

		// Wait for and process interrupts.

		for (;;)
		{
			// Wait for an interrupt to occur.
			sts	= WaitForRtInterrupt(irq_level, WAIT_FOREVER);

			if (dev->irq.os.do_exit)
			{
				dev->irq.os.told_to_exit	= 1;
				break;
			}

			if (sts == FALSE)
			{
				dev->irq.os.ret	= os_get_error_status();
				db_printf("%s: Failed wait for interrupt.\n", PREFIX);
				dev->irq.os.failed	= __LINE__;
				break;
			}

			// Service the interrupt.
			gsc_irq_isr_common(dev, GSC_IRQ_ISR_FLAG_LOCK);

			// Re-enable the interrupt then continue.
			sts	= EnableRtInterrupt(irq_level);

			if (sts == FALSE)
			{
				// The enable request failed.
				dev->irq.os.ret	= os_get_error_status();
				db_printf("%s: Failed to re-enable interrupt.\n", PREFIX);
				dev->irq.os.failed	= __LINE__;
				break;
			}
		}

		break;
	}

	dev->irq.os.stopped	= 1;

	sts	= DisableRtInterrupt(irq_level);

	if (sts == FALSE)
	{
		ret	= os_get_error_status();
		db_printf("%s: Failed to disable interrupt.\n", PREFIX);

		if (dev->irq.os.ret == 0)
			dev->irq.os.ret	= ret;

		if (dev->irq.os.failed == 0)
			dev->irq.os.failed	= __LINE__;
	}

	if (dev->irq.os.installed)
	{
		sts	= DeleteRtThread(NULL_RTHANDLE);

		if (sts == FALSE)
		{
			ret	= os_get_error_status();
			db_printf("%s: Failed to delete interrupt service thread.\n", PREFIX);

			if (dev->irq.os.ret == 0)
				dev->irq.os.ret	= ret;

			if (dev->irq.os.failed == 0)
				dev->irq.os.failed	= __LINE__;
		}

		sts	= ResetRtInterruptHandler(irq_level);

		if (sts == FALSE)
		{
			ret	= os_get_error_status();
			db_printf("%s: Failed to reset interrupt handler.\n", PREFIX);

			if (dev->irq.os.ret == 0)
				dev->irq.os.ret	= ret;

			if (dev->irq.os.failed == 0)
				dev->irq.os.failed	= __LINE__;
		}
	}
}



//*****************************************************************************
// Return:
//	0	Success
//	< 0	Failure: value is -errno or -GetLastRtError()
//	> 0	Never returned
static int _install_isr(dev_data_t* dev)
{
	int	ret;

	for (;;)	// A convenience loop
	{
		// Initialize all pertinent fields.
		dev->irq.os.started			= 0;
		dev->irq.os.stopped			= 0;
		dev->irq.os.ret				= 0;
		dev->irq.os.failed			= 0;
		dev->irq.os.installed		= 0;
		dev->irq.os.do_exit			= 0;
		dev->irq.os.told_to_exit	= 0;
		dev->irq.os.irq_level		= 0;
		dev->irq.os.thread			= 0;

		// Create the startup semaphore.
		ret	= os_sem_create_qty(1, 0, &dev->irq.os.sem);

		if (ret)
		{
			db_printf("%s: Failed to create startup semaphore.\n", PREFIX);
			dev->irq.os.failed	= __LINE__;
			break;
		}

		// Get the interrupt level.
		dev->irq.os.irq_level	= PciGetInterruptLevel(&dev->pci.pcidev);

		if (dev->irq.os.irq_level == 0xFFFF)
		{
			// This should never happen.
			ret	= -EINVAL;
			db_printf("%s: Failed to get interrupt level.\n", PREFIX);
			dev->irq.os.failed	= __LINE__;
			break;
		}

		// Start the interrupt service thread.
		dev->irq.os.thread	= CreateRtThread(255, _irq_thread, 8192, dev);

		if (dev->irq.os.thread == BAD_RTHANDLE)
		{
			// We couldn't create the thread.
			ret	= os_get_error_status();
			db_printf("%s: Failed to start interrupt service thread.\n", PREFIX);
			dev->irq.os.failed	= __LINE__;
			break;
		}

		// Wait for that thread to start working.
		ret	= os_sem_lock(&dev->irq.os.sem);

		if (ret)
		{
			db_printf("%s: Failed to acquire startup semaphore.\n", PREFIX);

			if (dev->irq.os.failed == 0)
				dev->irq.os.failed	= __LINE__;
		}

		break;
	}

	return(ret);
}



//*****************************************************************************
static void _remove_isr(dev_data_t* dev)
{
	int	ret;

	for (;;)	// A convenience loop
	{
		// Verify that we got the interrupt level.

		if ((dev->irq.os.irq_level == 0) ||
			(dev->irq.os.irq_level == 0xFFFF))
		{
			break;
		}

		// Verify that the service thread started.

		if ((dev->irq.os.thread == 0) ||
			(dev->irq.os.thread == BAD_RTHANDLE))
		{
			break;
		}

		// Terminate the interrupt service thread.
		dev->irq.os.do_exit	= 1;

		// We don't perform error checking here as there is the possibility
		// that the interrupt service thread has already exited. This can
		// occur if the threads starts running after we set the "quit" flag
		// and before be signal the thread.
		// Error checking was performed during development to validate
		SignalRtInterruptThread(dev->irq.os.irq_level);
		break;
	}

	ret	= os_sem_destroy(&dev->irq.os.sem);

	if (ret)
		db_printf("%s: Failed to destroy IRQ semaphore.\n", PREFIX);
}



//*****************************************************************************
// Return:
//	0	Success
//	< 0	Failure: value is -errno or -GetLastRtError()
//	> 0	Never returned
int os_irq_acquire(dev_data_t* dev)
{
	// The ISR is installed during the acquire call, so nothing is done here.
	return(0);
}



//*****************************************************************************
// Return:
//	0	Success
//	< 0	Failure: value is -errno or -GetLastRtError()
//	> 0	Never returned
int os_irq_create(dev_data_t* dev)
{
	int	ret;

	ret	= _install_isr(dev);
	return(ret);
}



//*****************************************************************************
void os_irq_destroy(dev_data_t* dev)
{
	_remove_isr(dev);
}



//*****************************************************************************
void os_irq_release(dev_data_t* dev)
{
	// The ISR is removed during the acquire call, so nothing is done here.
}



#endif
